import re

from pyinfra import logger
from pyinfra.api import FactBase

from .util.packaging import parse_packages

BREW_REGEX = r"^([^\s]+)\s([0-9\._+a-z\-]+)"


def new_cask_cli(version):
    """
    Returns true if brew is version 2.6.0 or later and thus has the new CLI for casks.
    i.e. we need to use brew list --cask instead of brew cask list
    See https://brew.sh/2020/12/01/homebrew-2.6.0/
    The version string returned by BrewVersion is a list of major, minor, patch version numbers
    """
    return (version[0] >= 3) or ((version[0] >= 2) and version[1] >= 6)


VERSION_MATCHER = re.compile(r"^Homebrew\s+(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*$")


def unknown_version():
    return [0, 0, 0]


class BrewVersion(FactBase):
    """
    Returns the version of brew installed as a semantic versioning tuple:

    .. code:: python

        [major, minor, patch]

    """

    command = "brew --version"
    requires_command = "brew"

    @staticmethod
    def default():
        return [0, 0, 0]

    def process(self, output):
        m = VERSION_MATCHER.match(output[0])
        if m is not None:
            return [int(m.group(key)) for key in ["major", "minor", "patch"]]
        logger.warning("could not parse version string from brew: %s", output[0])
        return self.default()


class BrewPackages(FactBase):
    """
    Returns a dict of installed brew packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "brew list --versions"
    requires_command = "brew"

    default = dict

    def process(self, output):
        return parse_packages(BREW_REGEX, output)


class BrewCasks(BrewPackages):
    """
    Returns a dict of installed brew casks:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = (
        r'if brew --version | grep -q -e "Homebrew\ +(1\.|2\.[0-5]).*" 1>/dev/null;'
        r"then brew cask list --versions; else brew list --cask --versions; fi"
    )
    requires_command = "brew"


class BrewTaps(FactBase):
    """
    Returns a list of brew taps.
    """

    command = "brew tap"
    requires_command = "brew"

    default = list

    def process(self, output):
        return output



from pyinfra.api import FactBase

from .util.packaging import parse_packages


class PkgPackages(FactBase):
    """
    Returns a dict of installed pkg packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    command = "pkg info || pkg_info || true"
    regex = r"^([a-zA-Z0-9_\-\+]+)\-([0-9a-z\.]+)"
    default = dict

    def process(self, output):
        return parse_packages(self.regex, output)



"""
Manage OpenVZ containers with ``vzctl``.
"""

from pyinfra import host
from pyinfra.api import OperationError, operation
from pyinfra.facts.vzctl import OpenvzContainers


@operation(is_idempotent=False)
def start(ctid, force=False):
    """
    Start OpenVZ containers.

    + ctid: CTID of the container to start
    + force: whether to force container start
    """

    args = ["{0}".format(ctid)]

    if force:
        args.append("--force")

    yield "vzctl start {0}".format(" ".join(args))


@operation(is_idempotent=False)
def stop(ctid):
    """
    Stop OpenVZ containers.

    + ctid: CTID of the container to stop
    """

    args = ["{0}".format(ctid)]

    yield "vzctl stop {0}".format(" ".join(args))


@operation(is_idempotent=False)
def restart(ctid, force=False):
    """
    Restart OpenVZ containers.

    + ctid: CTID of the container to restart
    + force: whether to force container start
    """

    yield from stop(ctid)
    yield from start(ctid, force=force)


@operation(is_idempotent=False)
def mount(ctid):
    """
    Mount OpenVZ container filesystems.

    + ctid: CTID of the container to mount
    """

    yield "vzctl mount {0}".format(ctid)


@operation(is_idempotent=False)
def unmount(ctid):
    """
    Unmount OpenVZ container filesystems.

    + ctid: CTID of the container to unmount
    """

    yield "vzctl umount {0}".format(ctid)


@operation(is_idempotent=False)
def delete(ctid):
    """
    Delete OpenVZ containers.

    + ctid: CTID of the container to delete
    """

    yield "vzctl delete {0}".format(ctid)


@operation(is_idempotent=False)
def create(ctid, template=None):
    """
    Create OpenVZ containers.

    + ctid: CTID of the container to create
    """

    # Check we don't already have a container with this CTID
    current_containers = host.get_fact(OpenvzContainers)
    if ctid in current_containers:
        raise OperationError(
            "An OpenVZ container with CTID {0} already exists".format(ctid),
        )

    args = ["{0}".format(ctid)]

    if template:
        args.append("--ostemplate {0}".format(template))

    yield "vzctl create {0}".format(" ".join(args))


@operation(is_idempotent=False)
def set(ctid, save=True, **settings):
    """
    Set OpenVZ container details.

    + ctid: CTID of the container to set
    + save: whether to save the changes
    + settings: settings/arguments to apply to the container

    Settings/arguments:
        these are mapped directly to ``vztctl`` arguments, eg
        ``hostname='my-host.net'`` becomes ``--hostname my-host.net``.
    """

    args = ["{0}".format(ctid)]

    if save:
        args.append("--save")

    for key, value in settings.items():
        # Handle list values (e.g. --nameserver X --nameserver X)
        if isinstance(value, list):
            args.extend("--{0} {1}".format(key, v) for v in value)
        else:
            args.append("--{0} {1}".format(key, value))

    yield "vzctl set {0}".format(" ".join(args))



import re

from pyinfra.api import FactBase, ShortFactBase


class Cpus(FactBase):
    """
    Returns the number of CPUs on this server.
    """

    command = "getconf NPROCESSORS_ONLN 2> /dev/null || getconf _NPROCESSORS_ONLN"

    @staticmethod
    def process(output):
        try:
            return int(output[0])
        except ValueError:
            pass


class Memory(FactBase):
    """
    Returns the memory installed in this server, in MB.
    """

    command = "vmstat -s"
    requires_command = "vmstat"

    @staticmethod
    def process(output):
        data = {}

        for line in output:
            line = line.strip()
            value, key = line.split(" ", 1)

            try:
                value = int(value)
            except ValueError:
                continue

            data[key.strip()] = value

        # Easy - Linux just gives us the number
        total_memory = data.get("K total memory", data.get("total memory"))

        # BSD - calculate the total from the # pages and the page size
        if not total_memory:
            bytes_per_page = data.get("bytes per page")
            pages_managed = data.get("pages managed")

            if bytes_per_page and pages_managed:
                total_memory = (pages_managed * bytes_per_page) / 1024

        if total_memory:
            return int(round(total_memory / 1024))


class BlockDevices(FactBase):
    """
    Returns a dict of (mounted) block devices:

    .. code:: python

        {
            "/dev/sda1": {
                "available": "39489508",
                "used_percent": "3",
                "mount": "/",
                "used": "836392",
                "blocks": "40325900"
            },
        }
    """

    command = "df"
    regex = r"([a-zA-Z0-9\/\-_]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([0-9]{1,3})%\s+([a-zA-Z\/0-9\-_]+)"  # noqa: E501
    default = dict

    def process(self, output):
        devices = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                if matches.group(1) == "none":
                    continue

                devices[matches.group(1)] = {
                    "blocks": matches.group(2),
                    "used": matches.group(3),
                    "available": matches.group(4),
                    "used_percent": matches.group(5),
                    "mount": matches.group(6),
                }

        return devices


class NetworkDevices(FactBase):
    """
    Gets & returns a dict of network devices. See the ``ipv4_addresses`` and
    ``ipv6_addresses`` facts for easier-to-use shortcuts to get device addresses.

    .. code:: python
        "enp1s0": {
            "ether": "12:34:56:78:9A:BC",
            "mtu": 1500,
            "state": "UP",
            "ipv4": {
                "address": "192.168.1.100",
                "mask_bits": 24,
                "netmask": "255.255.255.0"
            },
            "ipv6": {
                "address": "2001:db8:85a3::8a2e:370:7334",
                "mask_bits": 64,
                "additional_ips": [
                    {
                        "address": "fe80::1234:5678:9abc:def0",
                        "mask_bits": 64
                    }
                ]
            }
        },
        "incusbr0": {
            "ether": "DE:AD:BE:EF:CA:FE",
            "mtu": 1500,
            "state": "UP",
            "ipv4": {
                "address": "10.0.0.1",
                "mask_bits": 24,
                "netmask": "255.255.255.0"
            },
            "ipv6": {
                "address": "fe80::dead:beef:cafe:babe",
                "mask_bits": 64,
                "additional_ips": [
                    {
                        "address": "2001:db8:1234:5678::1",
                        "mask_bits": 64
                    }
                ]
            }
        },
        "lo": {
            "mtu": 65536,
            "state": "UP",
            "ipv6": {
                "address": "::1",
                "mask_bits": 128
            }
        },
        "veth98806fd6": {
            "ether": "AA:BB:CC:DD:EE:FF",
            "mtu": 1500,
            "state": "UP"
        },
        "vethda29df81": {
            "ether": "11:22:33:44:55:66",
            "mtu": 1500,
            "state": "UP"
        },
        "wlo1": {
            "ether": "77:88:99:AA:BB:CC",
            "mtu": 1500,
            "state": "UNKNOWN"
        }
    """

    command = "ip addr show 2> /dev/null || ifconfig -a"
    default = dict

    # Definition of valid interface names for Linux:
    # https://git.kernel.org/pub/scm/linux/kernel/git/stable/linux.git/tree/net/core/dev.c?h=v5.1.3#n1020
    def process(self, output):
        def mask(value):
            try:
                if value.startswith("0x"):
                    mask_bits = bin(int(value, 16)).count("1")
                else:
                    mask_bits = int(value)
                netmask = ".".join(
                    str((0xFFFFFFFF << (32 - b) >> mask_bits) & 0xFF) for b in (24, 16, 8, 0)
                )
            except ValueError:
                mask_bits = sum(bin(int(x)).count("1") for x in value.split("."))
                netmask = value

            return mask_bits, netmask

        # Strip lines and merge them as a block of text
        output = "\n".join(map(str.strip, output))

        # Splitting the output into sections per network device
        device_sections = re.split(r"\n(?=\d+: \w|\w+:.*mtu.*)", output)

        # Dictionary to hold all device information
        all_devices = {}

        for section in device_sections:
            # Extracting the device name
            device_name_match = re.match(r"^(?:\d+: )?([\w@]+):", section)
            if not device_name_match:
                continue
            device_name = device_name_match.group(1)

            # Regular expressions to match different parts of the output
            ether_re = re.compile(r"([0-9A-Fa-f:]{17})")
            mtu_re = re.compile(r"mtu (\d+)")
            ipv4_re = (
                re.compile(
                    r"inet (\d+\.\d+\.\d+\.\d+)/(\d+)(?: brd (\d+\.\d+\.\d+\.\d+))"
                ),  # ip a output,
                re.compile(
                    r"inet (\d+\.\d+\.\d+\.\d+)\s+netmask\s+((?:\d+\.\d+\.\d+\.\d+)|(?:[0-9a-fA-FxX]+))(?:\s+broadcast\s+(\d+\.\d+\.\d+\.\d+))"  # noqa: E501
                ),  # ifconfig -a output
            )

            # Parsing the output
            ether = ether_re.search(section)
            mtu = mtu_re.search(section)

            # Building the result dictionary for the device
            device_info = {}
            if ether:
                device_info["ether"] = ether.group(1)
            if mtu:
                device_info["mtu"] = int(mtu.group(1))

            device_info["state"] = (
                "UP" if "UP" in section else "DOWN" if "DOWN" in section else "UNKNOWN"
            )

            # IPv4 Addresses
            for ipv4_re_ in ipv4_re:
                ipv4_matches = ipv4_re_.findall(section)
                if ipv4_matches:
                    break

            if ipv4_matches:
                ipv4_info = []
                for ipv4 in ipv4_matches:
                    address = ipv4[0]
                    mask_value = ipv4[1]
                    mask_bits, netmask = mask(mask_value)
                    broadcast = ipv4[2] if len(ipv4) == 3 else None

                    ipv4_info.append(
                        {
                            "address": address,
                            "mask_bits": mask_bits,
                            "netmask": netmask,
                            "broadcast": broadcast,
                        },
                    )
                device_info["ipv4"] = ipv4_info[0]
                if len(ipv4_matches) > 1:
                    device_info["ipv4"]["additional_ips"] = ipv4_info[1:]

            # IPv6 Addresses
            ipv6_re = (
                re.compile(r"inet6\s+([0-9a-fA-F:]+)/(\d+)"),
                re.compile(r"inet6\s+([0-9a-fA-F:]+)\s+prefixlen\s+(\d+)"),
            )

            for ipv6_re_ in ipv6_re:
                ipv6_matches = ipv6_re_.findall(section)
                if ipv6_matches:
                    break

            if ipv6_matches:
                ipv6_info = []
                for ipv6 in ipv6_matches:
                    address = ipv6[0]
                    mask_bits = ipv6[1] or ipv6[2]
                    ipv6_info.append({"address": address, "mask_bits": int(mask_bits)})
                device_info["ipv6"] = ipv6_info[0]
                if len(ipv6_matches) > 1:
                    device_info["ipv6"]["additional_ips"] = ipv6_info[1:]

            all_devices[device_name] = device_info

        return all_devices


class Ipv4Addrs(ShortFactBase):
    """
    Gets & returns a dictionary of network interface -> list of IPv4 addresses.

    .. code:: python

        {
            "eth0": ["127.0.0.1"],
        }

    .. note::
        Network interfaces with no IPv4 will not be part of the dictionary.
    """

    fact = NetworkDevices
    ip_type = "ipv4"

    def process_data(self, data):
        host_to_ips = {}

        for interface, details in data.items():
            ips = []

            ip_details = details.get(self.ip_type)
            if not ip_details or not ip_details.get("address"):
                continue

            ips.append(ip_details["address"])
            if "additional_ips" in ip_details:
                ips.extend([ip["address"] for ip in ip_details["additional_ips"]])

            host_to_ips[interface] = ips

        return host_to_ips


class Ipv6Addrs(Ipv4Addrs):
    """
    Gets & returns a dictionary of network interface -> list of IPv6 addresses.

    .. code:: python

        {
            "eth0": ["fe80::a00:27ff::2"],
        }

    .. note::
        Network interfaces with no IPv6 will not be part of the dictionary.
    """

    ip_type = "ipv6"


# TODO: remove these in v3
# Legacy versions of the above that only support one IP per interface
#


class Ipv4Addresses(ShortFactBase):
    """
    Gets & returns a dictionary of network interface -> IPv4 address.

    .. code:: python

        {
            "eth0": "127.0.0.1",
        }

    .. warning::
        This fact is deprecated, please use the ``hardware.Ipv4Addrs`` fact.

    .. note::
        Network interfaces with no IPv4 will not be part of the dictionary.
    """

    fact = NetworkDevices
    ip_type = "ipv4"

    def process_data(self, data):
        addresses = {}

        for interface, details in data.items():
            ip_details = details.get(self.ip_type)
            if not ip_details or not ip_details.get("address"):
                continue  # pragma: no cover

            addresses[interface] = ip_details["address"]

        return addresses


class Ipv6Addresses(Ipv4Addresses):
    """
    Gets & returns a dictionary of network interface -> IPv6 address.

    .. code:: python

        {
            "eth0": "fe80::a00:27ff::2",
        }

    .. warning::
        This fact is deprecated, please use the ``hardware.Ipv6Addrs`` fact.

    .. note::
        Network interfaces with no IPv6 will not be part of the dictionary.
    """

    ip_type = "ipv6"



from pyinfra.api import FactBase

from .util.packaging import parse_packages

PIP_REGEX = r"^([a-zA-Z0-9_\-\+\.]+)==([0-9\.]+[a-z0-9\-]*)$"


class PipPackages(FactBase):
    """
    Returns a dict of installed pip packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    default = dict
    pip_command = "pip"

    def requires_command(self, pip=None):
        return pip or self.pip_command

    def command(self, pip=None):
        pip = pip or self.pip_command
        return "{0} freeze --all".format(pip)

    def process(self, output):
        return parse_packages(PIP_REGEX, output)


class Pip3Packages(PipPackages):
    pip_command = "pip3"



import os
from tempfile import mkstemp
from typing import TYPE_CHECKING, Optional

import click

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io, memoize
from pyinfra.progress import progress_spinner

from .local import run_shell_command as run_local_shell_command
from .util import make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class Meta(BaseConnectorMeta):
    handles_execution = True


@memoize
def show_warning():
    logger.warning("The @chroot connector is in beta!")


def make_names_data(directory: Optional[str] = None):
    if not directory:
        raise InventoryError("No directory provided!")

    show_warning()

    yield "@chroot/{0}".format(directory), {
        "chroot_directory": "/{0}".format(directory.lstrip("/")),
    }, ["@chroot"]


def connect(state: "State", host: "Host"):
    chroot_directory = host.data.chroot_directory

    try:
        with progress_spinner({"chroot run"}):
            local.shell(
                "chroot {0} ls".format(chroot_directory),
                splitlines=True,
            )
    except PyinfraError as e:
        raise ConnectError(e.args[0])

    host.connector_data["chroot_directory"] = chroot_directory
    return True


def run_shell_command(
    state: "State",
    host: "Host",
    command,
    get_pty: bool = False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output: bool = False,
    print_input: bool = False,
    return_combined_output: bool = False,
    **command_kwargs,
):
    chroot_directory = host.connector_data["chroot_directory"]

    command = make_unix_command_for_host(state, host, command, **command_kwargs)
    command = QuoteString(command)

    logger.debug("--> Running chroot command on (%s): %s", chroot_directory, command)

    chroot_command = StringCommand(
        "chroot",
        chroot_directory,
        "sh",
        "-c",
        command,
    )

    return run_local_shell_command(
        state,
        host,
        chroot_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state: "State",
    host: "Host",
    filename_or_io,
    remote_filename,
    remote_temp_filename=None,  # ignored
    print_output: bool = False,
    print_input: bool = False,
    **kwargs,  # ignored (sudo/etc)
):

    _, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, "wb") as temp_f:
                data = file_io.read()

                if isinstance(data, str):
                    data = data.encode()

                temp_f.write(data)

        chroot_directory = host.connector_data["chroot_directory"]

        chroot_command = "cp {0} {1}/{2}".format(
            temp_filename,
            chroot_directory,
            remote_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            chroot_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.remove(temp_filename)

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file uploaded to chroot: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status


def get_file(
    state: "State",
    host: "Host",
    remote_filename,
    filename_or_io,
    remote_temp_filename=None,  # ignored
    print_output: bool = False,
    print_input: bool = False,
    **kwargs,  # ignored (sudo/etc)
):

    _, temp_filename = mkstemp()

    try:
        chroot_directory = host.connector_data["chroot_directory"]
        chroot_command = "cp {0}/{1} {2}".format(
            chroot_directory,
            remote_filename,
            temp_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            chroot_command,
            print_output=print_output,
            print_input=print_input,
        )

        # Load the temporary file and write it to our file or IO object
        with open(temp_filename, encoding="utf-8") as temp_f:
            with get_file_io(filename_or_io, "wb") as file_io:
                data = temp_f.read()
                data_bytes: bytes

                if isinstance(data, str):
                    data_bytes = data.encode()
                else:
                    data_bytes = data

                file_io.write(data_bytes)
    finally:
        os.remove(temp_filename)

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file downloaded from chroot: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status



from unittest.mock import MagicMock, patch

from pyinfra.api import Config, State
from pyinfra.api.arguments import get_executor_kwarg_keys, pop_global_arguments
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.facts import get_facts
from pyinfra.facts.server import Arch, Command

from ..paramiko_util import PatchSSHTestCase
from ..util import make_inventory


def _get_executor_defaults(state, host):
    global_argument_defaults, _ = pop_global_arguments({}, state=state, host=host)
    return {
        key: value
        for key, value in global_argument_defaults.items()
        if key in get_executor_kwarg_keys()
    }


class TestFactsApi(PatchSSHTestCase):
    def test_get_fact(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(state, Command, ("yes",))

        assert fact_data == {anotherhost: "some-output"}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "yes",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost),
        )

    def test_get_fact_current_op_global_arguments(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")

        connect_all(state)
        anotherhost.current_op_global_kwargs = {
            "sudo": True,
            "sudo_user": "someuser",
            "use_sudo_password": True,
            "su_user": "someuser",
            "timeout": 10,
            "env": {"HELLO": "WORLD"},
        }

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(state, Command, ("yes",))

        assert fact_data == {anotherhost: "some-output"}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults.update(anotherhost.current_op_global_kwargs)

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "yes",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_fact_error(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = False, MagicMock()

            with self.assertRaises(PyinfraError) as context:
                get_facts(state, Command, ("fail command",))

        assert context.exception.args[0] == "No hosts remaining!"

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "fail command",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost),
        )

    def test_get_fact_error_ignore(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")

        connect_all(state)
        anotherhost.current_op_global_kwargs = {
            "ignore_errors": True,
        }

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = False, MagicMock()
            fact_data = get_facts(state, Command, ("fail command",))

        assert fact_data == {anotherhost: None}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "fail command",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **_get_executor_defaults(state, anotherhost),
        )

    def test_get_fact_executor_override_arguments(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(state, Command, ("yes",), {"_sudo": True})

        assert fact_data == {anotherhost: "some-output"}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults["sudo"] = True

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "yes",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_fact_executor_host_data_arguments(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")
        anotherhost.data._sudo = True

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(state, Command, ("yes",))

        assert fact_data == {anotherhost: "some-output"}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults["sudo"] = True

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "yes",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_fact_executor_mixed_arguments(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        anotherhost = inventory.get_host("anotherhost")
        anotherhost.data._sudo = True
        anotherhost.data._sudo_user = "this-should-be-overridden"
        anotherhost.data._su_user = "this-should-be-overridden"

        anotherhost.current_op_global_kwargs = {
            "su_user": "override-su-user",
        }

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(
                state,
                Command,
                args=("yes",),
                kwargs={"_sudo_user": "override-sudo-user"},
            )

        assert fact_data == {anotherhost: "some-output"}

        defaults = _get_executor_defaults(state, anotherhost)
        defaults["sudo"] = True
        defaults["sudo_user"] = "override-sudo-user"
        defaults["su_user"] = "override-su-user"

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            "yes",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_fact_cached(self):
        inventory = make_inventory(hosts=("anotherhost",))
        state = State(inventory, Config())

        fact_hash = "a-fact-hash"
        cached_fact = {"this is a cached fact"}
        anotherhost = inventory.get_host("anotherhost")
        anotherhost.facts[fact_hash] = cached_fact

        connect_all(state)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(
                state,
                Command,
                args=("yes",),
                kwargs={"_sudo": True},
                fact_hash=fact_hash,
            )

        assert fact_data == {anotherhost: cached_fact}
        fake_run_command.assert_not_called()

    def test_get_fact_no_args(self):
        inventory = make_inventory(hosts=("host-1",))
        state = State(inventory, Config())

        connect_all(state)

        host_1 = inventory.get_host("host-1")
        defaults = _get_executor_defaults(state, host_1)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = get_facts(state, Arch)

        assert fact_data == {host_1: "some-output"}
        fake_run_command.assert_called_with(
            state,
            host_1,
            Arch.command,
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )


class TestHostFactsApi(PatchSSHTestCase):
    def test_get_host_fact(self):
        inventory = make_inventory(hosts=("host-1",))
        state = State(inventory, Config())

        connect_all(state)

        host_1 = inventory.get_host("host-1")
        defaults = _get_executor_defaults(state, host_1)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = host_1.get_fact(Command, command="echo hello world")

        assert fact_data == "some-output"
        fake_run_command.assert_called_with(
            state,
            host_1,
            "echo hello world",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_host_fact_sudo(self):
        inventory = make_inventory(hosts=("host-1",))
        state = State(inventory, Config())

        connect_all(state)

        host_1 = inventory.get_host("host-1")
        defaults = _get_executor_defaults(state, host_1)
        defaults["sudo"] = True

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = host_1.get_fact(Command, command="echo hello world", _sudo=True)

        assert fact_data == "some-output"
        fake_run_command.assert_called_with(
            state,
            host_1,
            "echo hello world",
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )

    def test_get_host_fact_sudo_no_args(self):
        inventory = make_inventory(hosts=("host-1",))
        state = State(inventory, Config())

        connect_all(state)

        host_1 = inventory.get_host("host-1")
        defaults = _get_executor_defaults(state, host_1)
        defaults["sudo"] = True

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_run_command.return_value = MagicMock(), [("stdout", "some-output")]
            fact_data = host_1.get_fact(Arch, _sudo=True)

        assert fact_data == "some-output"
        fake_run_command.assert_called_with(
            state,
            host_1,
            Arch.command,
            print_input=False,
            print_output=False,
            return_combined_output=True,
            **defaults,
        )



import sys
from traceback import format_exception, format_tb

import click

from pyinfra import host, logger, state
from pyinfra.api.exceptions import PyinfraError
from pyinfra.context import ctx_host


class CliError(PyinfraError, click.ClickException):
    def show(self):
        name = "unknown error"

        if isinstance(self, PyinfraError):
            name = "pyinfra error"

        elif isinstance(self, IOError):
            name = "local IO error"

        if ctx_host.isset():
            # Get any operation meta + name
            op_name = None
            current_op_hash = host.current_op_hash
            current_op_meta = state.op_meta.get(current_op_hash)
            if current_op_meta:
                op_name = ", ".join(current_op_meta["names"])

            sys.stderr.write(
                "--> {0}{1}{2}: ".format(
                    host.print_prefix,
                    click.style(name, "red", bold=True),
                    " (operation={0})".format(op_name) if op_name else "",
                ),
            )
        else:
            sys.stderr.write(
                "--> {0}: ".format(click.style(name, "red", bold=True)),
            )

        logger.warning(self)


class UnexpectedMixin:
    def get_traceback_lines(self):
        traceback = getattr(self.e, "_traceback")
        return format_tb(traceback)

    def get_traceback(self):
        return "".join(self.get_traceback_lines())

    def get_exception(self):
        return "".join(format_exception(self.e.__class__, self.e, None))


class UnexpectedExternalError(click.ClickException, UnexpectedMixin):
    def __init__(self, e, filename):
        _, _, traceback = sys.exc_info()
        e._traceback = traceback
        self.e = e
        self.filename = filename

    def show(self):
        click.echo(
            "--> {0}:\n".format(
                click.style(
                    "An exception occurred in: {0}".format(self.filename),
                    "red",
                    bold=True,
                ),
            ),
            err=True,
        )

        click.echo(self.get_traceback(), err=True, nl=False)
        click.echo(self.get_exception(), err=True)


class UnexpectedInternalError(click.ClickException, UnexpectedMixin):
    def __init__(self, e):
        _, _, traceback = sys.exc_info()
        e._traceback = traceback
        self.e = e

    def show(self):
        click.echo(
            "--> {0}:\n".format(
                click.style(
                    "An internal exception occurred",
                    "red",
                    bold=True,
                ),
            ),
            err=True,
        )

        traceback_lines = self.get_traceback_lines()
        traceback = self.get_traceback()

        # Syntax errors contain the filename/line/etc, but other exceptions
        # don't, so print the *last* call to stderr.
        if not isinstance(self.e, SyntaxError):
            sys.stderr.write(traceback_lines[-1])

        exception = self.get_exception()
        click.echo(exception, err=True)

        with open("pyinfra-debug.log", "w", encoding="utf-8") as f:
            f.write(traceback)
            f.write(exception)

        logger.debug(traceback)
        logger.debug(exception)

        click.echo(
            "--> The full traceback has been written to {0}".format(
                click.style("pyinfra-debug.log", bold=True),
            ),
            err=True,
        )
        click.echo(
            (
                "--> If this is unexpected please consider submitting a bug report "
                "on GitHub, for more information run `pyinfra --support`."
            ),
            err=True,
        )



"""
Manage OpenRC init services.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.openrc import OpenrcEnabled, OpenrcStatus

from .util.service import handle_service_control


@operation
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
    runlevel="default",
):
    """
    Manage the state of OpenRC services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``rc-service <service> <command>``
    + enabled: whether this service should be enabled/disabled on boot
    + runlevel: runlevel to manage services for
    """

    yield from handle_service_control(
        host,
        service,
        host.get_fact(OpenrcStatus, runlevel=runlevel),
        "rc-service {0} {1}",
        running,
        restarted,
        reloaded,
        command,
    )

    if isinstance(enabled, bool):
        openrc_enabled = host.get_fact(OpenrcEnabled, runlevel=runlevel)
        is_enabled = openrc_enabled.get(service, False)

        if enabled and not is_enabled:
            yield "rc-update add {0}".format(service)
            openrc_enabled[service] = True

        if not enabled and is_enabled:
            yield "rc-update del {0}".format(service)
            openrc_enabled[service] = False



import json
from unittest import TestCase
from unittest.mock import patch

from pyinfra.api.exceptions import InventoryError
from pyinfra.connectors.terraform import make_names_data


class TestVagrantConnector(TestCase):
    def test_make_names_data_no_output_key(self):
        with self.assertRaises(InventoryError) as context:
            list(make_names_data())

        assert context.exception.args[0] == "No Terraform output key!"

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_no_output(self, fake_shell):
        fake_shell.return_value = json.dumps({})

        with self.assertRaises(InventoryError) as context:
            list(make_names_data("output_key"))

        assert context.exception.args[0] == "No Terraform output with key: `output_key`"

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_invalid_output(self, fake_shell):
        fake_shell.return_value = json.dumps({"output_key": "wrongvalue"})

        with self.assertRaises(InventoryError) as context:
            list(make_names_data("output_key"))

        assert (
            context.exception.args[0]
            == "Invalid Terraform output type, should be `list`, got `str`"
        )

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_dict_invalid_item(self, fake_shell):
        fake_shell.return_value = json.dumps({"output_key": [None]})

        with self.assertRaises(InventoryError) as context:
            list(make_names_data("output_key"))

        assert (
            context.exception.args[0]
            == "Invalid Terraform list item, should be `dict` or `str` got `NoneType`"
        )

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data(self, fake_shell):
        fake_shell.return_value = json.dumps({"output_key": ["somehost"]})
        data = list(make_names_data("output_key"))

        assert data == [
            (
                "@terraform/somehost",
                {"ssh_hostname": "somehost"},
                ["@terraform"],
            ),
        ]

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_nested(self, fake_shell):
        fake_shell.return_value = json.dumps({"output_key": {"nested_key": ["somehost"]}})
        data = list(make_names_data("output_key.nested_key"))

        assert data == [
            (
                "@terraform/somehost",
                {"ssh_hostname": "somehost"},
                ["@terraform"],
            ),
        ]

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_dict(self, fake_shell):
        host = {
            "name": "a name",
            "ssh_hostname": "hostname",
        }
        fake_shell.return_value = json.dumps({"output_key": [host]})
        data = list(make_names_data("output_key"))

        assert data == [
            (
                "@terraform/a name",
                {"ssh_hostname": "hostname"},
                ["@terraform"],
            ),
        ]

    @patch("pyinfra.connectors.terraform.local.shell")
    def test_make_names_data_dict_no_name(self, fake_shell):
        host = {
            "not_a_name": "hostname",
        }
        fake_shell.return_value = json.dumps({"output_key": [host]})

        with self.assertRaises(InventoryError) as context:
            list(make_names_data("output_key"))

        assert (
            context.exception.args[0]
            == "Invalid Terraform list item, missing `name` or `ssh_hostname` keys"
        )



"""
Manage XBPS packages and repositories. Note that XBPS package names are case-sensitive.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.xbps import XbpsPackages

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def upgrade():
    """
    Upgrades all XBPS packages.
    """

    yield "xbps-install -y -u"


_upgrade = upgrade  # noqa: E305


@operation(is_idempotent=False)
def update():
    """
    Update XBPS repositories.
    """

    yield "xbps-install -S"


_update = update  # noqa: E305


@operation
def packages(
    packages=None,
    present=True,
    update=False,
    upgrade=False,
):
    """
    Install/remove/update XBPS packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + update: run ``xbps-install -S`` before installing packages
    + upgrade: run ``xbps-install -y -u`` before installing packages

    **Example:**

    .. code:: python

        xbps.packages(
            name="Install Vim and Vim Pager",
            packages=["vimpager", "vim"],
        )

    """

    if update:
        yield from _update()

    if upgrade:
        yield from _upgrade()

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(XbpsPackages),
        present,
        install_command="xbps-install -y -u",
        uninstall_command="xbps-remove -y",
    )



from urllib.parse import urlparse

from pyinfra.api import FactBase


class GpgFactBase(FactBase):
    abstract = True

    requires_command = "gpg"

    key_record_type = "pub"
    subkey_record_type = "sub"

    def process(self, output):
        # For details on the field values see:
        # https://git.gnupg.org/cgi-bin/gitweb.cgi?p=gnupg.git;a=blob_plain;f=doc/DETAILS
        keys = []

        current_key = None
        current_subkey = None

        for line in output:
            if not line:
                continue

            bits = line.split(":")

            if bits[0] in (self.key_record_type, self.subkey_record_type):
                key_details = {
                    "validity": bits[1],
                    "length": int(bits[2]),
                    "id": bits[4],
                }

                if bits[0] == self.key_record_type:
                    key_details["subkeys"] = []

                    if current_key:
                        if current_subkey:
                            current_key["subkeys"].append(current_subkey)

                        keys.append(current_key)

                    current_key = key_details
                    current_subkey = None
                elif current_key:
                    current_subkey = key_details

            elif current_subkey or current_key:
                target = current_subkey or current_key

                if bits[0] == "fpr":
                    target["fingerprint"] = bits[9]  # fingerprint = field 10
                elif bits[0] == "uid":
                    target["uid_hash"] = bits[7]
                    target["uid"] = bits[9]

        if current_key:
            if current_subkey:
                current_key["subkeys"].append(current_subkey)

            keys.append(current_key)

        # Return as a dictionary of keyID -> details
        keys_by_id = {}
        for key in keys:
            if "subkeys" in key:
                key["subkeys"] = {subkey.pop("id"): subkey for subkey in key["subkeys"]}
            keys_by_id[key.pop("id")] = key

        return keys_by_id


class GpgKey(GpgFactBase):
    """
    Returns information on one or more GPG keys found in a file or URL.

    .. code:: python

        {
            "KEY-ID": {
                "length": 4096,
                "uid": "Oxygem <hello@oxygem.com>"
            },
        }
    """

    def command(self, src):
        if urlparse(src).scheme:
            return ("(wget -O - {0} || curl -sSLf {0}) | gpg --with-colons").format(src)

        return "gpg --with-colons {0}".format(src)


class GpgKeys(GpgFactBase):
    """
    Returns information on all public keys in a keychain.

    .. code:: python

        {
            "KEY-ID": {
                "length": 4096,
                "uid": "Oxygem <hello@oxygem.com>"
            },
        }
    """

    def command(self, keyring=None):
        if not keyring:
            return "gpg --list-keys --with-colons"

        return ("gpg --list-keys --with-colons --keyring {0} --no-default-keyring").format(keyring)


class GpgSecretKeys(GpgFactBase):
    """
    Returns information on all secret keys in a keychain.

    .. code:: python

        {
            "KEY-ID": {
                "length": 4096,
                "fingerprint": "ABC",
                "uid": "Oxygem <hello@oxygem.com>"
            },
        }
    """

    key_record_type = "sec"
    subkey_record_type = "ssb"

    def command(self, keyring=None):
        if not keyring:
            return "gpg --list-secret-keys --with-colons"

        return ("gpg --list-secret-keys --with-colons --keyring {0} --no-default-keyring").format(
            keyring,
        )
